# -*- coding: utf-8 -*-

# Import from Django
from django.contrib.admin import ModelAdmin
from django.contrib import admin
from django import forms
from django.utils.translation import ugettext_lazy as _

from ckeditor.widgets import CKEditorWidget
from modeltranslation.admin import TranslationAdmin

from crma.models import (
    MailingList,
    Channel,
    Email,
    EmailScheduler,
    Subscription,
    Contact,
)
from crma.forms import MailingListAdminForm


class EmailForm(forms.ModelForm):
    class Meta:
        model = Email
        widgets = {
            'text': CKEditorWidget(),
        }
        fields = '__all__'


class EmailAdmin(TranslationAdmin):
    form = EmailForm
    list_display = [
        'channel',
        'created',
        'subject',
        'interval',
        'tag',
        'enabled',
    ]
    list_filter = [
        'channel',
        'enabled',
    ]
    date_hierarchy = 'created'

    class Media:
        js = [
            'modeltranslation/js/force_jquery.js',
            'js/jquery-ui.js',
            'modeltranslation/js/tabbed_translation_fields.js',
        ]
        css = {
            'screen': [
                'modeltranslation/css/tabbed_translation_fields.css',
            ],
        }


class EmailSchedulerAdmin(ModelAdmin):
    list_display = [
        'from_address',
        'contact',
        'email',
        'lang',
        'ctime',
        'sched_time',
        'sent_time',
        'status',
    ]
    search_fields = [
        'from_address',
        'email__subject',
        'contact__email',
    ]
    list_filter = [
        'status',
        'lang',
        'email',
    ]
    readonly_fields = [
        'extra_context',
        'trace_error',
    ]

    fieldsets = [
        (_('Emails'), {
            'fields': (
                'from_address', 'contact', 'email', 'lang', 'sched_time'
            )}),
        (_('Status'), {
            'fields': (
                ('sent_time', 'status'),
            )}),
        (_('Context'), {
            'fields': (
                ('extra_context', 'trace_error'),
            )}),
    ]


class SubscriptionAdmin(ModelAdmin):
    list_display = [
        'channel',
        'contact',
        'state',
    ]
    search_fields = [
        'channel__title',
        'contact__email',
    ]
    list_filter = [
        'channel',
        'state',
    ]
    fieldsets = [
        (_('Emails'), {
            'fields': (
                'channel', 'contact',
            )}),
        (_('Subscription info'), {
            'fields': (
                ('state', 'unsubscribe_key'),
            )}),
    ]
    readonly_fields = [
        'unsubscribe_key',
    ]


class ChannelAdmin(TranslationAdmin):
    list_display = [
        'title',
        'from_address',
        'required',
    ]
    list_filter = [
        'title',
        'from_address',
        'required',
    ]
    prepopulated_fields = {
        'channel_id': ['title']
    }

    class Media:
        js = [
            'modeltranslation/js/force_jquery.js',
            'js/jquery-ui.js',
            'modeltranslation/js/tabbed_translation_fields.js',
        ]
        css = {
            'screen': [
                'modeltranslation/css/tabbed_translation_fields.css',
            ],
        }


class ContactAdmin(ModelAdmin):
    list_display = [
        'email',
        'lang',
    ]
    list_filter = [
        'lang',
    ]
    search_fields = [
        'email',
    ]


class MembersInline(admin.TabularInline):
    model = MailingList.members.through
    extra = 0
    raw_id_fields = [
        'contact',
    ]


class MailingListAdmin(ModelAdmin):
    list_display = [
        'title',
    ]
    filter_horizontal = [
        'members',
    ]
    form = MailingListAdminForm
    exclude = [
        'members',
    ]

    inlines = [
        MembersInline,
    ]


try:
    from import_export.admin import ImportExportMixin
except ImportError:
    pass
else:
    from crma.resources import ContactResource
    ContactAdmin = type(ContactAdmin)('ContactAdmin', (ImportExportMixin, ContactAdmin,), {
        'resource_class': ContactResource,
    })

admin.site.register(Channel, ChannelAdmin)
admin.site.register(Email, EmailAdmin)
admin.site.register(EmailScheduler, EmailSchedulerAdmin)
admin.site.register(Subscription, SubscriptionAdmin)
admin.site.register(Contact, ContactAdmin)
admin.site.register(MailingList, MailingListAdmin)
